/*
 * squash.c
 *		Ver 1.0 (2015/09/25)
 *			ncurses 版
 *		Ver 1.1 (2016/09/23)
 *			ncursesw 版
 *		Ver 1.2 (2016/10/14)
 *			ncursesw 版
 */

/*
 *	必要なライブラリ : libncursesw5-dev
 *		sudo apt-get install libncursesw5-dev
 */

/* ncurses 版
 * #include <curses.h>
 */

#include <ncursesw/ncurses.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>

/*
 *	乱数関係
 */

void init_random() {

  srand((unsigned)time(NULL));	/* 乱数を現在時刻で初期化 */
}

int nrand(int n) {		/* n>1 とし 0 〜 n-1 の整数の乱数を生成する */
  int r = rand();

  if ( r < 0 ) {
	r = -r;
  }

  return r % n;
}

int brand(int min, int max)	{
	/* min < max とし min 〜 max の乱数を生成する */

  return min + nrand(max-min+1);
}

/*
 * スクリーンサイズ
 */

#define	SCREEN_X_SIZE	79		/* スクーン X サイズ */
#define	SCREEN_Y_SIZE	23		/* スクーン Y サイズ */

#define	SCREEN_X_MIN	0
#define	SCREEN_X_MAX	(SCREEN_X_SIZE-1)

#define	SCREEN_Y_MIN	0
#define	SCREEN_Y_MAX	(SCREEN_Y_SIZE-1)

#define	SCREEN_X_SEPARATE	60

/*
 * ゲーム画面サイズ
 */

#define GAME_X_MIN	(SCREEN_X_MIN+1)
#define GAME_Y_MIN	(SCREEN_Y_MIN+1)

#define GAME_X_MAX	(SCREEN_X_SEPARATE-1)
#define GAME_Y_MAX	(SCREEN_Y_MAX-1)

/*
 * ラケット情報
 */

#define	Y_RACKET	(SCREEN_Y_MAX -2 )	/* ラケットの位置 */
#define	RACKET_LEN	8		/* ラケットの長さ */

/*
 * 時間情報
 */

#define	W_MSEC	1000		/* 各操作の最小単位時間(小さい程速い) */
#define	W_COUNT	100			/* ボールの移動までの時間 */

/*
 * ゲーム開始メッセージ
 */

#define	GAME_START_X	(SCREEN_X_SIZE/3)
#define	GAME_START_Y	(SCREEN_Y_SIZE/3)

int	game_start() {
  int ch;

  mvaddstr ( GAME_START_Y, GAME_START_X, "Game Start ? (y=yes/q=quit)" );
  refresh();					/* 書き換えた内容を画面に描画 */

  while ( ( ch = getch() ) != 'y' ) {  			/* 文字の入力 */
	if ( ch == 'q' ) {
	  break;
	}
  }

  return ch;
}

/*
 * ゲームオーバーメッセージ
 */

#define	GAME_OVER_X	(SCREEN_X_SIZE/3)
#define	GAME_OVER_Y	(SCREEN_Y_SIZE/3)
#define	GAME_OVER_WAIT (1000000 * 2)		/* 2 秒待つ */

void	game_over() {

  mvaddstr ( GAME_START_Y, GAME_START_X, "Game Over" );
  refresh();					/* 書き換えた内容を画面に描画 */

  usleep ( GAME_OVER_WAIT );

}

/*
 *	draw_frame
 *		枠を描画する
 *
 */

void draw_fram() {
  int x;
  int y;

  for ( x = SCREEN_X_MIN + 1; x <= SCREEN_X_MAX - 1; x++ ) {
	mvaddch ( SCREEN_Y_MIN, x, '-' );
	mvaddch ( SCREEN_Y_MAX, x, '-' );
  }

  for ( y = SCREEN_Y_MIN + 1; y <= SCREEN_Y_MAX - 1; y++ ) {
	mvaddch ( y, SCREEN_X_MIN, '|' );
	mvaddch ( y, SCREEN_X_SEPARATE, '|' );
	mvaddch ( y, SCREEN_X_MAX, '|' );
  }

  mvaddch ( SCREEN_Y_MIN, SCREEN_X_MIN, '+' );
  mvaddch ( SCREEN_Y_MIN, SCREEN_X_SEPARATE, '+' );
  mvaddch ( SCREEN_Y_MIN, SCREEN_X_MAX, '+' );

  mvaddch ( SCREEN_Y_MAX, SCREEN_X_MIN, '+' );
  mvaddch ( SCREEN_Y_MAX, SCREEN_X_SEPARATE, '+' );
  mvaddch ( SCREEN_Y_MAX, SCREEN_X_MAX, '+' );

}

/*
 *	draw_number ( int y, int x, int number, int digits )
 *		指定した桁数で数値を出力する
 */

void draw_number ( int y, int x, int number, int digits ) {

  while ( digits > 0 ) {
	mvaddch ( y, x + digits - 1, '0' + (number % 10) );
	number /= 10;	/* number = number / 10;	*/
	digits--;		/* digit = digit - 1; 		*/
  }

}

/*
 *	draw_ball_postion
 */

#define	STATUS_BALL_X	(SCREEN_X_SEPARATE+2)
#define	STATUS_BALL_Y	5

void draw_ball_postion ( int bx, int by ) {

  mvaddstr ( STATUS_BALL_Y, STATUS_BALL_X, "Ball (   ,    )" );
  draw_number ( STATUS_BALL_Y, STATUS_BALL_X +  7, bx, 2 );
  draw_number ( STATUS_BALL_Y, STATUS_BALL_X + 11, by, 2 );

}

/*
 *	draw_score
 */

#define	STATUS_SCORE_X	(SCREEN_X_SEPARATE+2)
#define	STATUS_SCORE_Y	(STATUS_BALL_Y+1)

void draw_score ( int score ) {

  mvaddstr ( STATUS_SCORE_Y, STATUS_SCORE_X, "Score: " );
  draw_number (STATUS_SCORE_Y, STATUS_SCORE_X  +  7, score, 8 );

}

/*
 *	draw_max_score
 */

#define	STATUS_MAX_SCORE_X	(SCREEN_X_SEPARATE+2)
#define	STATUS_MAX_SCORE_Y	(STATUS_SCORE_Y+1)

void draw_max_score ( int score ) {

  mvaddstr ( STATUS_MAX_SCORE_Y, STATUS_MAX_SCORE_X, "Max:   " );
  draw_number (STATUS_MAX_SCORE_Y, STATUS_MAX_SCORE_X  +  7, score, 8 );

}

/*
 *	draw_status
 *		色々な情報を提供する
 *
 */

#define	STATUS_RACKET_X	(SCREEN_X_SEPARATE+2)
#define	STATUS_RACKET_Y	6

void draw_status (
				  int bx, int by,
				  int score,
				  int max_score
				  ) {

  draw_ball_postion ( bx, by );
  draw_score ( score );
  draw_max_score ( max_score );

}

/*
 * ブロック
 */

#define	BLOCK_Y_LENGTH	3
#define	BLOCK_X_LENGTH	15

#define	BLOCK_Y_POS		3
#define	BLOCK_X_POS		10

void draw_block ( int block[BLOCK_Y_LENGTH][BLOCK_X_LENGTH] ) {
  int x;
  int y;

  /* ブロックの初期変化 */

  for ( y = 0; y < BLOCK_Y_LENGTH; y++ ) {
	  for ( x = 0; x < BLOCK_X_LENGTH; x++ ) {
		if ( block[y][x] == 1 ) {
		  move ( BLOCK_Y_POS + y, BLOCK_X_POS + x * 3 );
		  addch ( '%' );
		}
	  }
  }

}

/*
 *	report
 *		ボールやラケットの位置を表示する
 */

void report ( int bx, int by, int rx ) {
  move ( 1, GAME_X_MAX + 1 );
  refresh();
  printf ( "bx=%d, by=%d rx=%d", bx, by, rx );
}

/*
 * erase_ball
 *		ボールを消す
 */

void erase_ball ( int bx, int by ) {
	move ( by, bx );	/* ボールの位置 (bx,by) にカーソルを移動 */
	addch ( ' ' );		/* ' ' を上書きする事により、ボールを消す */
}

/*
 * draw_ball
 *		ボールを書く
 */

void draw_ball ( int bx, int by ) {
	move ( by, bx );	/* ボールの位置 (bx,by) にカーソルを移動 */
	addch ( '*' );		/* '*' を上書きする事により、ボールを書く */
}

/*
 * erase_racket
 *		ラケットを消す
 */

void erase_racket ( int rx ) {
  int i;

  /* ラケットの長さだけ */
  for ( i = 0; i < RACKET_LEN; i++ ) {
		/* ラケットを作る文字の位置にカーソルを移動 */
	move ( Y_RACKET, rx - RACKET_LEN / 2 + i );
		/* ラケットを消す */
	addch ( ' ' );
  }
}

/*
 * erase_racket
 *		ラケットを書く
 */

void draw_racket ( int rx ) {
  int i;

  /* ラケットの長さだけ */
  for ( i = 0; i < RACKET_LEN; i++ ) {
		/* ラケットを作る文字の位置にカーソルを移動 */
	move ( Y_RACKET, rx - RACKET_LEN / 2 + i );
		/* ラケットを書く */
	addch ( '=' );
  }
}

/*
 * squash_game
 *	スカッシュゲーム本体
 */


int squash_game( int max_score ) {
  int ch;
  int bx = GAME_X_MAX/2;		/* ボールの位置の x 座標 (初期値は中央) */
  int by = GAME_Y_MAX/2;		/* ボールの位置の y 座標 (初期値は中央) */
  int dbx = brand(-2,2);	/* ボールの速度の x 成分 (初期値は 乱数) */
  int dby = -1;			/* ボールの速度の y 成分 (初期値は -1 [上向き] ) */
  int rx = GAME_X_MAX/2;		/* ラケット中央の位置の x 座標 (初期値は中央) */
  int count = 0;		/* ボールが動くまでの待機時間 */
  int score = 0;		/* ゲームのスコア */

  int block[BLOCK_Y_LENGTH][BLOCK_X_LENGTH];	/* ブロック */
  int x;
  int y;

  /* ブロックの初期変化 */

  for ( y = 0; y < BLOCK_Y_LENGTH; y++ ) {
	  for ( x = 0; x < BLOCK_X_LENGTH; x++ ) {
		block[y][x] = 1;
	  }
  }

  /* ゲーム開始 */

  erase();				/* 画面を全て消す */
  draw_fram();			/* 枠を書く */

  while ( 0 == 0 ) {		/* 無限 Loop */

	usleep ( W_MSEC );		/* W_MSEC だけ待つ (最小時間を一つ進める) */

	count ++;				/* ボールの待機時刻を進める */

	if ( count > W_COUNT ) {	/* ボールの移動時刻が来た */
	  count = 0;				/* 待機時間時刻を最初に戻す */
	  /* report ( bx, by, rx ); */

	  erase_ball( bx, by );		/* 現在あるボールを削除 */

	  /* ボールの x 座標の処理 */

	  bx = bx + dbx;			/* ボールの x 座標を更新 */
	  							/* 「速度」だけ増やす */

	  if ( bx < GAME_X_MIN ) {		/* 画面の左端に来た */
		dbx = 1;					/* 「速度」の向きを右向きにする */
		bx = GAME_X_MIN + 1;
	  }

	  if ( bx > GAME_X_MAX ) {		/* 画面の右端に来た */
		dbx = -1;					/* 「速度」の向きを左向きにする */
		bx = GAME_X_MAX - 1;
	  }

	  /* ボールの y 座標の処理 */

	  by = by + dby;			/* ボールの y 座標を更新 */

	  if ( by < GAME_Y_MIN ) {	/* 画面の上端に来た */
		dby = 1;				/* 「速度」の向きを下向きにする */
		by = GAME_Y_MIN + 1;

		if ( nrand(4) > 0 ) {	/* 乱数で、x 方向の速度を変更する */
		  bx = bx - dbx;		/* 元に戻し */
  		  dbx = brand(-2,2);	/* 速度を変更し */
		  bx = bx + dbx;		/* その位置に移動 */
		}

		score++;				/* score = score + 1; スコアを増やす */
		if ( score > max_score ) {
		  max_score = score;	/* 最大スコアを更新 */
		}
	  }

	  if ( by > GAME_Y_MAX ) {		/* 画面の下端に来た */
		//		dby = -1;				/* 「速度」の向きを上向きにする */
		//	by = GAME_Y_MAX - 1;
		break;					/* ゲームオーバ */
	  }

	  /* ボールとブロック衝突チェック */

	  for ( y = 0; y < BLOCK_Y_LENGTH; y++ ) {
		for ( x = 0; x < BLOCK_X_LENGTH; x++ ) {
		  if ( block[y][x] == 1 ) {
			if ( by == BLOCK_Y_POS + y && bx == BLOCK_X_POS + x * 3 ) {
			  block[y][x] = 0;
			  dbx = - dbx;
			  dby = - dby;
			}
		}
	  }
  }

	  /* ボールとラケット衝突チェック */

	  /* 衝突した */
	  if ( by == Y_RACKET && rx - RACKET_LEN/2 <= bx && bx < rx + RACKET_LEN/2 ){
		dby = -1;	/* ボールの速度の向きを「上向き」にする */
		by = Y_RACKET - 2;
	  }

	  draw_ball( bx, by );	/* ボールを表示する */
	}

	/*
	 * 入力処理 (ラケットの移動)
	 */	 

	ch = getch();  			/* 文字の入力 */

	erase_racket ( rx );	/* ラケットを消す */

	/* 入力の内容によって、ラケットの位置を移動 */

	if ( ch == 'h' ) {			/* 'h' (左方向) が入力された */
	  if ( rx - RACKET_LEN/2 > GAME_X_MIN ) {	/* 左端でなければ */
		rx--;					/* 左に移動 */
	  }
	} else if ( ch == 'l' ) {	/* 'l' (右方向) が入力された */
	  if ( rx + RACKET_LEN/2 <= GAME_X_MAX ) {	/* 右端でなければ */
		rx++;					/* 右に移動 */
	  }
	} else if ( ch == 'q' ) {	/* 'q' (ゲーム終了) が入力された */
	  score = -1;				/* スコアを -1 にして中断である事を示す */
	  break;					/* while loop を抜ける(ゲーム終了) */
	}

	draw_block ( block );

	draw_racket ( rx );			/* 新しい位置にラケットを表示する */

	draw_status ( bx, by, score, max_score );

	refresh();					/* 書き換えた内容を画面に描画 */
  }

  return score;					/* スコアを返す */

}

/*
 * main
 */

int main(void) {
  int max_score = 0;
  int score;

  /* 乱数の初期化 */

  init_random();

  /* 画面の初期化 */

  initscr();		/* 画面を初期化 (クリアされる) */
  cbreak();			/* 文字の入力をリアルタイムにする */
  noecho();			/* 文字の入力をエコーバックしない */

  timeout(0);		/* 文字の入力がなければ直ぐにかえる */

  while ( game_start() != 'q'  ) {

	if ( (score = squash_game( max_score ) ) < 0 )	{	/* スコアが 0 以下なら無条件に終了 */
	  break;		/* ゲームを完全に終了 */
	} else {
	  if ( score > max_score ) {			/* 最大スコアのチェック */
		max_score = score;
	  }
	  game_over();	/* ゲームオーバー */
	}
  }

  endwin();			/* curses の後始末 */

  return 0;
}
